/***************************************************************************
                                                    SIMPLY POWERFUL TOOLKIT (SPTK)
                                                    CIntMap.h  -  description
                                                          -------------------
        begin                : 13 June 2003
        copyright            : (C) 2000-2003 by Alexey S.Parshin
        email                : alexeyp@gmail.com
  ***************************************************************************/

/***************************************************************************
   This library is free software; you can redistribute it and/or modify it 
   under the terms of the GNU Library General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at 
   your option) any later version.

   This library is distributed in the hope that it will be useful, but 
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library
   General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; if not, write to the Free Software Foundation, 
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.

   Please report all bugs and problems to "alexeyp@gmail.com"
 ***************************************************************************/

#ifndef __CINTMAP_H__
#define __CINTMAP_H__

#include <stdlib.h>
#include <string.h>

#include <sptk/CList.h>
//#include <sptk/CString.h>

/// Class CBaseIntMapItem represents an item for CBaseIntMap.
class CBaseIntMapItem {
public:
	int   m_id;    ///< Item ID
	void *m_data;  ///< Item data - a pointer to the user object
public:
	/// Contructor.
	/// @param id int, the item ID
	/// @param data void *, data
	CBaseIntMapItem(int id, void *data) { m_id = id; m_data = data; }
};

class CBaseIntMapIterator;

/// Class CBaseIntMap is a base class for CIntMap. 
/// It implements most of the functionality of the CIntMap.
class CBaseIntMap {
	friend class CBaseIntMapIterator;
protected:
   /// The internal list of the items
	CList   m_items;

   /// The internal list search function
	static int integersSearch(const void *key, const void *item);

   /// Finds an item by the item ID
	CBaseIntMapItem *findItem(int id) const;

   /// Inserts the item with the id. If the item with such id already exists,
   /// it is replaced with the new one
	void            *insertItem(int id,void *data);

   /// Destroys the user object
	virtual void     deleteData(void *data) {}

public:

	//// Default constructor.
	CBaseIntMap() {}

	/// Destructor
	virtual ~CBaseIntMap();

public:

   /// Removes all the map items.
	virtual void clear();

   /// Reports if the map has no items.
	/// @returns true if map has no items.
	bool empty() const { return m_items.count() == 0; }

   /// Reports the number of items in the map.
	/// @returns number of items in the map.
	unsigned count() const { return m_items.count(); }

	/// Direct item access.
	/// @param pos unsigned, item position number
	/// @returns an item in position
	void *item(unsigned pos) const { return m_items[pos]; }

   /** Inserts an item with ID. If the item with same ID
    * already exists in the map, it will be replaced
    * @param id int, the item ID
    * @param data void *, item data
    */
	void *insert(int id,void *data);

   /** Removes an item (if any) with the ID.
    * @param id int, item ID
    */
	void remove(int id);

   /** Finds an item (if any) with the ID.
    * @param id int, item ID
    * @returns map item, or NULL if not found
    */
	void *find(int id) const;

   /** Throws an exception if item isn't found. Internal. 
    * @param id int, item ID
    */
	void throwNotFoundException(int id) const;
};

/// An iterator for CIntMap. 
/// Allows to go through the map items sequentually.
class CBaseIntMapIterator {
	const CBaseIntMap *m_map;       ///< A pointer to map to work with
	unsigned           m_position;  ///< Current position in the map
public:

	/// Constructor.
	/// @param map CBaseIntMap, map to work with.
	CBaseIntMapIterator(const CBaseIntMap& map);

	/// Moves the current position to the first item 
	/// @returns map item, or NULL if not found
	void *first();

	/// Moves the current position to the last item 
	/// @returns map item, or NULL if not found
	void *last();

	/// Returns the item in the current position to the first item 
	/// @returns map item, or NULL if outside the map
	void *operator *  ();

	/// Increments the current position in the map 
	/// @returns map item, or NULL if outside the map
	void *operator ++ (int);

	/// Decrements the current position in the map 
	/// @returns map item, or NULL if outside the map
	void *operator -- (int);

	/// Returns the integer id of the item in the map
	int ident() const;
};

/// Class CIntMap is a template class that implements a map of integer
/// IDs to your objects. 
template <class T>
class CIntMap : public CBaseIntMap {
protected:   
   /// Destroys the user object
	virtual void    deleteData(void *data) { delete (T *) data; }
public:

   /// Constructor
	CIntMap() {}

   /// Destructor. Calls clear() to remove all the items. 
	virtual ~CIntMap() { 
      clear(); 
   }

   /** Item access (const version) by item ID.  
    * @param id int, item ID
    */
	const T& operator [] (int id) const  {
		void *ptr = CBaseIntMap::find(id);
		if (!ptr)
			throwNotFoundException(id);
		return *(T *)ptr;
	}

   /** Item access (non-const version) by item ID.  
    * @param id int, item ID
    */
	T& operator [] (int id) {
		void *ptr = CBaseIntMap::find(id);
		if (!ptr)
			throwNotFoundException(id);
		return *(T *)ptr;
	}

   /** Item access (const version) by item index.  
    * @param index int, item position
    */
	T& item(int index) const                         { return *(T *)CBaseIntMap::item(index); }

   /** Finds item by ID.  
    * @param id int, item ID
    * @returns item pointer, or NULL if item not found
    */
	T* find(int id) const                          { return (T *)CBaseIntMap::find(id);   }

   /** Inserts new item into the map. If an item with the same already exists in the map
    * it is destroyed and replaced with the new one.  
    * @param id int, item ID
    * @param item T *, an item pointer.
    */
	T *insert(int id,T *item)                    { return (T *)CBaseIntMap::insert(id,item);        }

   /** Inserts new item into the map. If an item with the same already exists in the map
    * it is destroyed and replaced with the copy of new one.  
    * @param id int, item ID
    * @param item T *, an item pointer.
    */
	T *insert(int id,const T& item) { 
      T* existing = find(id);
      if (existing) {
         *existing = item;
         return existing;
      }
      return (T *)CBaseIntMap::insert(id,new T(item)); 
   }

   /// Removes the item from the list
   /// @param id int, item ID
	void remove(int id)                            { CBaseIntMap::remove(id);             }

   /// Finds out if the map contains an item
   /// @param id int, item ID
   /// @returns true if there is such item in the map
	bool contains(int id) const                    { return find(id) != 0L;  }

};

/** Class CIntMapIterator is a template for CIntMap iterator. */
template <class T>
class CIntMapIterator {
   /// Actual iterator
	CBaseIntMapIterator  itor;
public:
   /// Constructor 
   /// @param map const CIntMap<T>&, map to work with
	CIntMapIterator(const CIntMap<T>& map) : itor(map) {}

   /// Constructor 
   /// @param map const CIntMap<T>*, map to work with
	CIntMapIterator(const CIntMap<T>* map) : itor(*map) {}

   /// Moves the current position to the first item 
	/// @returns map item, or NULL if not found
	T* first()                                     { return (T *) itor.first(); }

	/// Moves the current position to the last item 
	/// @returns map item, or NULL if not found
	T* last()                                      { return (T *) itor.first(); }

	/// Increments the current position in the map 
	/// @returns map item, or NULL if outside the map
	T* operator ++ (int)                           { return (T *) itor++;       }

	/// Decrements the current position in the map 
	/// @returns map item, or NULL if outside the map
	T* operator -- (int)                           { return (T *) itor--;       }

	/// Returns the item in the current position to the first item 
	/// @returns map item, or NULL if outside the map
	T* operator *  ()                              { return (T *) *itor;        }

	/// Returns the integer id of the item in the map
	int ident() const                              { return itor.ident();       }
};

#endif
