/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CList.h  -  description
                             -------------------
    begin                : Fri May 26 2000
    copyright            : (C) 2000-2003 by Alexey S.Parshin
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
   This library is free software; you can redistribute it and/or modify it 
   under the terms of the GNU Library General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at 
   your option) any later version.

   This library is distributed in the hope that it will be useful, but 
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library
   General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; if not, write to the Free Software Foundation, 
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.

   Please report all bugs and problems to "alexeyp@gmail.com"
 ***************************************************************************/

#ifndef __CLIST_H__
#define __CLIST_H__

#include <sptk/sptk.h>

typedef void *          PListItem;

typedef int (*CForeachFunction)(void *item,void *arg);

/// List of the pointers. Designed mostly to be a part of other widgets
/// and to be a base class for CObjectList, etc. Uses effective strategy
/// of forward memory allocation.
class SP_EXPORT CList {

   unsigned    m_capacity;   ///< The number of elements in the list
   unsigned    m_size;       ///< The maximum number of elements in the list without memory re-allocation

protected:

   PListItem   *m_items;     ///< The elements array

public:

   /// Default constructor
   /// @param sz unsigned, the initial size of the list. The items in the list are not initialized.
   CList(unsigned sz=0);

   /// Destructor
   ~CList();

   /// Empty the list
   void clear() { resize(0); }

   /// Resize the list to hold at least newSize elements.
   /// The excessive memory is released.
   void resize(unsigned newSize);

   /// Adds a new element into the list
   /// @param item PListItem, a pointer
   void add(PListItem item);

   /// Adds another list of pointers into the list
   /// @param itemList const CList&, a list of pointers to add
   void addList(const CList& itemList);

   /// Inserts a new element into the list in the position pos
   /// @param pos unsigned, insert position
   /// @param item PListItem, a pointer
   void insert(unsigned pos,PListItem item);

   /// Removes an element from the list in the position pos
   /// @param pos unsigned, insert position
   void remove(unsigned pos);

   /// Removes an element from the list in the position pos
   /// @param item PListItem, a list element 
   bool remove(PListItem item);

   /// Removes an elements from the list which are also in itemList
   /// @param itemList const CList&, a list of elements to remove
   bool removeList(const CList& itemList);

   /// Sorts an elements of the list with fcmp function
   /// @param fcmp a sort function
   void sort(int (*fcmp)(const void *, const void *));

   /// Number of elements in the list
   unsigned count() const { return m_size; }

   /// Element access by index, non-const version
   /// @param ind unsigned, element index
   /// @returns item of the list
   PListItem& operator [] (unsigned ind) { return m_items[ind]; }

   /// Element access by index, const version
   /// @param ind unsigned, element index
   /// @returns item of the list
   const PListItem operator [] (unsigned ind) const { return m_items[ind]; }

   /// Finds an index of an element in the list
   /// @param item const void *, a list element 
   /// @returns an index of the element, or -1 if not found
   int indexOf(const void * item) const;

   /// Returns internal items' array
   const void *base() const { return (const void *)m_items; }

   /// Calls the todo function for every element of the array
   /// till it returns true. Non-const version.
   int foreach(CForeachFunction todo,void * arg);

   /// Calls the todo function for every element of the array
   /// till it returns true. Const version.
   int foreach(CForeachFunction todo,void * arg) const;
};

#endif
