/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CObjectList.h  -  description
                             -------------------
    begin                : Jul 1 2003
    copyright            : (C) 2000-2003 by Alexey S.Parshin
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
   This library is free software; you can redistribute it and/or modify it 
   under the terms of the GNU Library General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at 
   your option) any later version.

   This library is distributed in the hope that it will be useful, but 
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library
   General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; if not, write to the Free Software Foundation, 
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.

   Please report all bugs and problems to "alexeyp@gmail.com"
 ***************************************************************************/

#ifndef __COBJECTLIST_H__
#define __COBJECTLIST_H__

#include <sptk/CList.h>
#include <sptk/CException.h>

/** Class CObjectList is the template for the objects list. */

template <class T>
class SP_EXPORT CObjectList {
protected:
   /// Internal list of items
   CList          m_items;
   /// Flag to maintain items memory
   bool           m_autoCleanup;
public:

   /** Constructor
    * @param autoCleanup bool, true if the list should maintain the items' memory
    */
   CObjectList(bool autoCleanup=true)        { m_autoCleanup = autoCleanup; }

   /** Copy constructor. Data items are copied using the 
     * copy constructor of template type.
     */
   CObjectList(const CObjectList& ol)        { 
      m_autoCleanup = ol.m_autoCleanup; 
      unsigned cnt = ol.count();
      for (unsigned i = 0; i < cnt; i++) {
         add(*(T*)ol.m_items[i]);
      }
   }

   /** Destructor */
   ~CObjectList()                            { clear(); }

   /** Adds an item to the list.
     * The list will contain a copy of the pointer, not the independent copy of the object.
     * @param item T *, item pointer
     */
   void add(T *item)                         { m_items.add(item); }

   /** Adds an item to the list. 
     * The list will contain the independent from item copy of the object.
     * The object better have a copy constructer for this to work.
     * @param item const T&, item
     */
   void add(const T& item)                   { m_items.add(new T(item)); }

   /** Inserts an item into the list.
     * The list will contain a copy of the pointer, not the independent copy of the object.
     * @param pos unsigned, item position to insert in
     * @param item T *, item pointer
     */
   void insert(unsigned pos,T *item)         { m_items.insert(pos,item); }

   /** Inserts an item into the list. 
     * The list will contain the independent from item copy of the object.
     * The object better have a copy constructer for this to work.
     * @param pos unsigned, item position to insert in
     * @param item const T&, item
     */
   void insert(unsigned pos,const T& item)   { m_items.insert(pos,(void *)new T(item)); }

   /** Removes the item in position pos. Depending
     * on parameter autoCleanup in constructor, 
     * the object is optionally deleted first 
     * @param pos unsigned, item position to remove from
     */
   void remove(unsigned pos) {
      if (m_autoCleanup)
         delete (T *) m_items[pos];
      m_items.remove(pos);
   }

   /** Removes the item with address ptr. Depending
     * on parameter autoCleanup in constructor, 
     * the object is optionally deleted first 
     * @param ptr T *, item pointer
     */
   bool remove(T *ptr) {
      if (m_autoCleanup)
         delete ptr;
      return m_items.remove(ptr);
   }

   /** Removes all the items from the list. Depending
     * on parameter autoCleanup in constructor, 
     * the objects are optionally deleted first 
     */
   void clear() {
      if (m_autoCleanup) {
         unsigned cnt = m_items.count();
         for (unsigned i = 0; i < cnt; i++)
            delete (T *) m_items[i];
      }
      m_items.clear();
   }

   /** Resizes (truncates or extends) the list. Depending
     * on parameter autoCleanup in constructor, 
     * the objects are optionally deleted/created
     */
   void resize(unsigned newsize) {
      unsigned cnt = m_items.count();
      if (newsize > cnt)
         m_items.resize(newsize);
      if (m_autoCleanup) {
         unsigned i;
         for (i = newsize; i < cnt; i++)
            delete (T *) m_items[i];
         for (i = cnt; i < newsize; i++)
            m_items[i] = new T();
      }
      if (newsize < cnt)
         m_items.resize(newsize);
   }

   /** List item access by index */
   T& operator [] (unsigned ind)             { return *(T *)m_items[ind]; }

   /** List item access by index */
   const T& operator [] (unsigned ind) const { return *(T *)m_items[ind]; }

   /** List item access by index */
   T& operator [] (int ind)                  { return *(T *)m_items[ind]; }

   /** List item access by index */
   const T& operator [] (int ind) const      { return *(T *)m_items[ind]; }

   /** Sorts the list using the sort function */
   void  sort(int (*sort_function)(const void *, const void *)) {
      m_items.sort(sort_function);
   }

   /** Copies data from other list. Data items are copied using the 
     * copy constructor of template type.
     */
   CObjectList& operator = (const CObjectList& n) {
      clear();
      unsigned cnt = n.count();
      for (unsigned i = 0; i < cnt; i++) {
         add(*(T*)n.m_items[i]);
      }
      return *this;
   }

   /** Reports number of items in the list */
   unsigned count() const              { return m_items.count(); }

   /** Finds the item position in the list
    * @param p T *, item address
    * @returns item position (if found) or -1 (if not found)
    */
   int indexOf(const T* p) const       { return m_items.indexOf(p); }

   /** Finds the item position in the list
    * @param p T& , item reference
    * @returns item position (if found) or -1 (if not found)
    */
   int indexOf(const T& p) const       { return m_items.indexOf(&p); }

   /** Moves the item one position up
    * @param p const T*, item address
    * @returns true if completed, or false if the item is already at the start of the list 
    */
   bool moveUp(const T *p) {
      int itemPos = indexOf(p);
      if (itemPos < 1) return false;
      void *priorItem = m_items[itemPos-1];
      m_items[itemPos] = priorItem;
      m_items[itemPos-1] = (void *)p;
      return true;
   }

   /** Moves the item one position down
    * @param p const T*, item address
    * @returns true if completed, or false if the item is already at the end of the list 
    */
   bool moveDown(const T *p) {
      unsigned itemPos = indexOf(p);
      if (itemPos + 2 > m_items.count()) return false;
      void *nextItem = m_items[itemPos+1];
      m_items[itemPos] = nextItem;
      m_items[itemPos+1] = (void *)p;
      return true;
   }

   /** Stack class top() emulation */
   T& top() {
      int ndx = m_items.count() - 1;
      if (ndx < 0)
         throw CException("Can't get top() - the stack is empty.");
      return *(T *)m_items[ndx];
   }

   /** Stack class push() emulation */
   void push(const T& _item) {
      m_items.add((void *)new T(_item));
   }

   /** Stack class pop() emulation */
   T pop() {
      int ndx = m_items.count() - 1;
      if (ndx < 0)
         throw CException("Can't pop() - the stack is empty.");

      T tempCopy = *(T *)m_items[ndx];
      m_items.remove(ndx);

      return tempCopy;
   }
};
#endif
