/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CStringList.h  -  description
                             -------------------
    begin                : Tue Jan 11 2000
    copyright            : (C) 2000-2003 by Alexey S.Parshin
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
   This library is free software; you can redistribute it and/or modify it 
   under the terms of the GNU Library General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at 
   your option) any later version.

   This library is distributed in the hope that it will be useful, but 
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library
   General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; if not, write to the Free Software Foundation, 
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.

   Please report all bugs and problems to "alexeyp@gmail.com"
 ***************************************************************************/

#ifndef __CSTRINGLIST_H__
#define __CSTRINGLIST_H__

#include <sptk/CList.h>
#include <sptk/CString.h>
#include <sptk/CBuffer.h>

class CPackedStrings;

	/// Extended version of CString that supports an integer string ID
class SP_EXPORT CStringInfo : public CString {
public:
   /// String ID
	int       id;
public:
   /// Default constructor
	CStringInfo();
   /// Copy constructor
	CStringInfo(const CStringInfo& si);
   /// Constructor
   /// @param str const CString&, source string
	CStringInfo(const CString& str);
   /// Constructor
   /// @param str const char *, source string
	CStringInfo(const char *str);
   /// Constructor
   /// @param str const char *, source string
   /// @param ident int, string ID
   /// @param maxlen int, maximum number of characters to copy
	CStringInfo(const char *str,int ident,int maxlen=-1);

   /// Assignment operator
	CStringInfo& operator = (const CStringInfo& si);
   /// Assignment operator
	CStringInfo& operator = (const CString& si);
   /// Assignment operator
	CStringInfo& operator = (const char * s);

   /// Returns string ID
	int   ident() const { return id; }
};

typedef CStringInfo *   PCStringInfo;

/// The list od strings. Strings are represented as CStringInfo objects.
/// Every string in the list may have an ID.
class CStringList {
protected:
   /// The list of strings
	CList  m_list;
   /// User data for the list
	void  *m_userData;
public:

   /// Constructor
   /// @param sz unsigned, the starting list size
	CStringList(unsigned sz=0);

   /// Copy constructor
	CStringList(const CStringList& sl);

   /// Constructor
   /// @param csvString const char *, string with tokens separated with delimitter
   /// @param delimitter CString, the delimitter
	CStringList(const char *csvString,CString delimitter);

   /// Constructor
   /// @param csvString const CString&, string with tokens separated with delimitter
   /// @param delimitter CString, the delimitter
	CStringList(const CString& csvString,CString delimitter);

   /// Constructor
   /// @param csvBuffer const CBuffer&, buffer with tokens separated with delimitter
   /// @param delimitter CString, the delimitter
	CStringList(const CBuffer& csvBuffer,CString delimitter);

   /// Constructor
   /// @param strings const CPackedStrings&, packed strings 
	CStringList(const CPackedStrings& strings);

   /// Constructor
	~CStringList();

   /// Returns number of strings in the list
	unsigned count() const { return m_list.count(); }

   /// Resizes the list
   /// @param sz unsigned, the new list size
	void resize(unsigned sz);

   /// String access by index, non-const version
   /// @param index unsigned, string index in the list
	CStringInfo& operator[] (unsigned index) { return *(CStringInfo *)m_list[index]; }

   /// String access by index, const version
   /// @param index unsigned, string index in the list
	const CStringInfo& operator[] (unsigned index) const { return *(CStringInfo *)m_list[index]; }

   /// Assignment operator
	CStringList& operator =(const CStringList& sl); // ASSIGNMENT OPERATOR.

   /// Adds string into the list
   /// @param s const char *, the string to insert
	void add(const char *s)                      { m_list.add(new CStringInfo(s)); }

   /// Adds string into the list
   /// @param s const char *, the string to insert
   /// @param id int, string ID
   /// @param maxlen int, maximum characters to copy
	void add(const char *s,int id,int maxlen=-1) { m_list.add(new CStringInfo(s,id,maxlen)); }

   /// Adds string into the list
   /// @param s const CString&, the string to insert
	void add(const CString& s)                   { m_list.add(new CStringInfo(s)); }

   /// Adds string into the list
   /// @param si const CStringInfo&, the string to insert
	void add(const CStringInfo& si)              { m_list.add(new CStringInfo(si)); }

   /// Adds strings into the list
   /// @param sl const CStringList&, the strings to insert
	void add(const CStringList& sl);

   /// Inserts string into the list
   /// @param index unsigned, the insert position
   /// @param s const char *, the string to insert
   /// @param id int, string ID
	void insert(unsigned index,const char *s,int id = 0);

   /// Inserts string into the list
   /// @param index unsigned, the insert position
   /// @param si const CStringInfo&, the string to insert
	void insert(unsigned index,const CStringInfo& si);

   /// Inserts strings into the list
   /// @param index unsigned, the insert position
   /// @param sl const CStringList&, the strings to insert
	void insert(unsigned index,const CStringList& sl);

   /// Removes string from the list
   /// @param index unsigned, the string index in the list
	void remove(unsigned index);

   /// Removes all the strings from the list
	void clear() { resize(0); }

   /// Finds a string
   /// @param s const char *, the string to find
   /// @returns string index in the list, or -1 if not found
	int  indexOf(const char *s) const;

   /// Finds a string
   /// @param s const CString&, the string to find
   /// @returns string index in the list, or -1 if not found
	int  indexOf(const CString& s) const { return indexOf(s.c_str()); }

   /// Finds a string
   /// @param id int, the string ID
   /// @returns string index in the list, or -1 if not found
	int  indexOf(int id) const;

   /// Sorts strings in the list
	void sort();

   /// Returns string containing strings from the list separated with delimitter
   /// @param delimitter const char *, the delimitter
	CString asString(const char *delimitter=",") const;

   /// Sets string list from tokens in string separated with delimitter
   /// @param csvString const CString&, string with tokens separated with delimitter
   /// @param delimitter CString, the delimitter
	void fromString(const CString& csvString,const char *delimitter=",");

   /// Sets string list from tokens in buffer separated with delimitter
   /// @param csvBuffer const CBuffer&, zero-terminated string with tokens separated with delimitter
   /// @param delimitter CString, the delimitter
	void fromBuffer(const CBuffer& csvBuffer,const char *delimitter=",");

   /// Saves strings to file
   /// @param fileName CString, the name of the file
	bool saveToFile(CString fileName) const;

   /// Loads strings from file
   /// @param fileName CString, the name of the file
	bool loadFromFile(CString fileName);

   /// Returns user data as integer
	int   argument() const   { return (int)m_userData; }

   /// Sets user data as integer
	void  argument(int d)    { m_userData = (void *)d; }

   /// Returns user data as void *
	void *user_data() const  { return m_userData;      }

   /// Sets user data as void *
	void  user_data(void *d) { m_userData = d;         }

   /// Add operation (list concatenation)
   /// @returns this object as a result of concatenation
	CStringList& operator += (const CStringList& list);

   /// Sub operation (list substruction)
   /// Be careful, this operation would work slow on a big lists.
   /// @returns new CStringList as a result of substruction
	CStringList& operator -= (const CStringList& list);
};
//---------------------------------------------------------------------------
#endif
