/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CStringMap.h  -  description
                             -------------------
    begin                : 13 June 2003
    copyright            : (C) 2000-2003 by Alexey S.Parshin
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
   This library is free software; you can redistribute it and/or modify it 
   under the terms of the GNU Library General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at 
   your option) any later version.

   This library is distributed in the hope that it will be useful, but 
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library
   General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; if not, write to the Free Software Foundation, 
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.

   Please report all bugs and problems to "alexeyp@gmail.com"
 ***************************************************************************/

#ifndef __CSTRINGMAP_H__
#define __CSTRINGMAP_H__

#include <stdlib.h>
#include <string.h>

#include <sptk/CList.h>
#include <sptk/CString.h>

/// Class CBaseStringMapItem represents a single item in the map.
class CBaseStringMapItem {
public:
   /// Item name, unique in the map
	char *m_name;  
   /// Item data
	void *m_data;
public:
   /// Constructor.
   /// @param name const char *, item name
   /// @param data void *, item data
	CBaseStringMapItem(const char *name, void *data);

   /// Destructor.
	~CBaseStringMapItem();
};

class CBaseStringMapIterator;

/// Class CBaseStringMap is a base class for CStringMap.
class CBaseStringMap {
	friend class CBaseStringMapIterator;
protected:
   /// Internal list of the map items
	CList   m_items;

   /// Map mode for the keys
	bool    m_ignoreCase;

   /// Callback item name search function.
	static int stringsSearch(const void *key, const void *item);

   /// Finds an item index by the item name.
   /// @param name const char *, item name
   /// @returns item index, or -1
	int findItemIndex(const char *name) const;

   /// Finds an item by the item name.
   /// @param name const char *, item name
   /// @returns item data pointer, or NULL
	CBaseStringMapItem *findItem(const char *name) const;

   /// Inserts an item for the item name. 
   /// If the item with such name already exists it is replaced.
   /// @param name const char *, item name
   /// @param data void *, item data
   /// @returns new item pointer
	void           *insertItem(const char *name,void *data);

   /// Deletes an item in the derived class
	virtual void    deleteData(void *data) {}
public:

   /// Default constructor
   /// @param ignoreCase bool, true if map should ignore the case for the map keys
	CBaseStringMap(bool ignoreCase);

   /// Destructor
	virtual ~CBaseStringMap();

   /// Removes all items. Items memory is maintained in the derived class.
	virtual void clear();

   /// Reports true if the map has no items
	bool empty() const { return m_items.count() == 0; }

   /// Returns number of itens in the map
	unsigned count() const { return m_items.count(); }

   /// Returns an item in the map by the item index
	void *item(unsigned index) const { return ((CBaseStringMapItem *)m_items[index])->m_data; }

   /// Inserts an item in the map
   /// @param name const char *, item name
   /// @param data void *, item
   /// @returns new item pointer
	void *insert(const char *name,void *data);

   /// Inserts an item in the map
   /// @param name const CString&, item name
   /// @param data void *, item
   /// @returns new item pointer
	void *insert(const CString& name,void *data);

   /// Removes an from the map
   /// @param name const char *, item name
	void remove(const char *name);

   /// Removes an from the map
   /// @param name const CString&, item name
	void remove(const CString& name);

   /// Finds an item in the map
   /// @param name const char *, item name
   /// @returns item pointer, or NULL if not found
	void *find(const char *name) const;

   /// Throws "item not found" exception
	void throwNotFoundException(const char *name) const;

   /// Returns map mode for the keys
	bool ignoreCase() const { return m_ignoreCase; }
};

/// Class CBaseStringMapIterator is the base class for CStringMapIterator.
/// It implements most of the functionality of the derived class.
class CBaseStringMapIterator {
	const CBaseStringMap  *m_map;       ///< Map to work with
	unsigned               m_position;  ///< Position in the map
public:
   /// Constructor
   /// @param map const CBaseStringMap&, map to work with
	CBaseStringMapIterator(const CBaseStringMap& map);
   /// Moves map pointer to the beginning
	void *first();
   /// Moves map pointer to the end
	void *last();
   /// Returns the pointer on the current item, or NULL if outside the map
	void *operator *  ();
   /// Moves to the next item
	void *operator ++ (int);
   /// Moves to the prior item
	void *operator -- (int);
   /// Returns current item's name
	const char *name() const;
	/// Returns true if pointer is inside the map
	bool valid() { return m_position < m_map->count(); }
};

/// Template class CStringMap implements a map of strings to user objects.
template <class T>
class CStringMap : public CBaseStringMap {

protected:

   /// deletes the map item
	virtual void    deleteData(void *data) { 
		delete (T *) data; 
	}

public:

   /// Default constructor
   /// @param ignoreCase bool, true if map should ignore the case for the map keys
	CStringMap(bool ignoreCase=false) : CBaseStringMap(ignoreCase) {}

   /// Destructor.
   /// Deletes all the map items.
	virtual ~CStringMap() { clear(); }

   /// Deletes all the map items.
	void clear() {
		unsigned cnt = m_items.count();
		for (unsigned i = 0; i < cnt; i++) {
			CBaseStringMapItem *currentItem = (CBaseStringMapItem *) m_items[i];
			delete (T *)currentItem->m_data;
			delete currentItem;
		}
		m_items.clear(); 
	}

   /// Item access by name, const version
	const T& operator [] (const char *name) const  {
		void *ptr = CBaseStringMap::find(name);
		if (!ptr)
			throwNotFoundException(name);
		return *(T *)ptr;
	}

   /// Item access by name, non-const version
	T& operator [] (const char *name) {
		void *ptr = CBaseStringMap::find(name);
		if (!ptr)
			throwNotFoundException(name);
		return *(T *)ptr;
	}

   /// Item access by index, const version
	T&   item(int index) const                     { return *(T *)CBaseStringMap::item(index); }

   /// Finds item access by index, const version
	T*   find(const char *name) const              { return (T *)CBaseStringMap::find(name);   }

   /// Inserts an item into the map, without creating an item copy.
   /// Such items are still deleted in the map's destructor.
   /// @param name const char *, item name
   /// @param item <T *>, item pointer
   /// @returns new item pointer
	T*   insert(const char *name,T *item)          { return (T *)CBaseStringMap::insert(name,item);        }

   /// Inserts an item copy into the map
   /// @param name const char *, item name
   /// @param item <const T &>, item reference
   /// @returns new item pointer
	T*   insert(const char *name,const T& item)    { 
		T* existing = (T *)CBaseStringMap::find(name);
		if (existing) {
			*existing = item;
			return existing;
		}
		return (T *)CBaseStringMap::insert(name,new T(item)); 
	}

   /// Finds out if the map contains an item
   /// @param name const char *, item name
   /// @returns true if there is such item in the map
	bool contains(const char *name) const          { return CBaseStringMap::find(name) != 0L;  }
};

/// Template class CStringMapIterator designed to work with CStringMap class objects.
template <class T>
class CStringMapIterator : public CBaseStringMapIterator {
public:

   /// Constructor
   /// @param map const CStringMap<T>&, map to work with
	CStringMapIterator(const CStringMap<T>& map) : CBaseStringMapIterator(map) {}

   /// Constructor
   /// @param map const CStringMap<T>*, map to work with
	CStringMapIterator(const CStringMap<T>* map) : CBaseStringMapIterator(*map) {}

   /// Moves iterator to the beginning of the map
   /// @returns pointer to the current item, or 0L if there are no items in the map
	T* first()                                     { return (T *) CBaseStringMapIterator::first();        }

   /// Moves iterator to the end of the map
   /// @returns pointer to the current item, or 0L if there are no items in the map
	T* last()                                      { return (T *) CBaseStringMapIterator::first();        }

   /// Moves iterator to the next item of the map
   /// @returns pointer to the current item, or 0L if there are no more items in the map to iterate
	T* operator ++ (int)                           { return (T *) CBaseStringMapIterator::operator++(0);  }

   /// Moves iterator to the prior item of the map
   /// @returns pointer to the current item, or 0L if there are no more items in the map to iterate
	T* operator -- (int)                           { return (T *) CBaseStringMapIterator::operator--(0);  }

   /// Returns current item of the map
   /// @returns pointer to the current item, or 0L if there are no more items in the map to iterate
	T* operator *  ()                              { return (T *) CBaseStringMapIterator::operator *();   }

   /// Returns current item's value, non-const version
	T& value()                                     { return *(T *) CBaseStringMapIterator::operator *();  }

   /// Returns current item's value, const version
	const T& value() const                         { return *(T *) CBaseStringMapIterator::operator *();  }
};

#endif
