/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CXmlDoc.h  -  description
                             -------------------
    begin                : Sun May 22 2003
    based on the code    : Mikko Lahteenmaki <Laza@Flashmail.com>
    copyright            : (C) 2003 by Alexey S.Parshin
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
   This library is free software; you can redistribute it and/or modify it 
   under the terms of the GNU Library General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at 
   your option) any later version.

   This library is distributed in the hope that it will be useful, but 
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library
   General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; if not, write to the Free Software Foundation, 
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.

   Please report all bugs and problems to "alexeyp@gmail.com"
 ***************************************************************************/

#ifndef __CXMLDOC_H__
#define __CXMLDOC_H__

#include "CXmlNode.h"
#include "CXmlTokenizer.h"

#include <sptk/CBuffer.h>

/// @addtogroup XML
/// @{

/// Maps an XML entity string to a presentation string.
class CXmlEntities : public CStringMap<CString> {
public:
   typedef CStringMapIterator<CString> Iterator;   ///< Entities iterator

   /// Constructor
   CXmlEntities() { }

    /**
     * Remove named entity.
     * @param name const char *, entity name to remove
     * Returns true, if operation success
     */
   void removeEntity(const char *name) { remove(name); }

    /**
     * Add entity to map, If entity named 'name' exists already in map,
     * it's value is replaced with 'replacement'
     * @param name const char *, entity to add/change
     * @param replacement const char *, value that represents entity
     */
   void setEntity(const char *name, const char *replacement) { insert(name,replacement); }
};

/**
 * The CXmlDocType represents tag <DOCTYPE ...> in XML document.
 * It can return a map of all entities(). 
 * This class provides the name(), public_id() and system_id() functions. 
 */
class CXmlDocType {
   friend class CXmlParser;
public:

   /// Constructor
   CXmlDocType() { m_html = false; }
   /// Constructor
   CXmlDocType(const char *name, const char *public_id=0, const char *system_id=0);

    /**
     * Returns the name of the document type as specified in the <!DOCTYPE name> tag.
     */
   const CString &name() const { return m_name; }

    /**
     * Returns the public identifier of the external DTD subset.
     * Returns empty string if there is no public identifier.
     */
   const CString &publicID() const { return m_public_id; }

    /**
     * Returns the system identifier of the external DTD subset.
     * Returns empty string if there is no system identifier.
     */
   const CString &systemID() const { return m_system_id; }

   /**
    * Returns a map of all entities described in the DTD.
    * NOTE: Map doesn't hold default entities.
    */
   CXmlEntities &entities() { return m_entities;  }

   /**
    * Returns a map of all entities described in the DTD.
    * NOTE: Map doesn't hold default entities.
    */
   const CXmlEntities &entities() const { return m_entities;  }

    /**
     * Encodes string to XML representation.
     * I.e. Converts "<test>" to "&lt;test&gt;"
     * Returns true, any entities replaced.
     * @param str const char *, string to convert
     * @param ret CString &, converted string is stored to this.
     */
   bool encodeEntities(const char *str, CString &ret);

    /**
     * Decodes entities in string to their actual values.
     * I.e. Converts "&lt;test&gt;" to "<test>"
     * Returns true, any entities expanded.
     * @param str string to convert
     * @param ret converted string is stored to this.
     */
   bool decodeEntities(const char *str, CString &ret);

    /**
     * Search for entity with given name
     * Returns true, if attribute is found.
     * @param name const char *, entity to search
     */
   bool hasEntity(const char *name) const;

    /**
     * Remove named entity from entity map.
     * Returns true, if entity removed.
     * @param name entity to remove
     */
   void removeEntity(const char *name) { m_entities.removeEntity(name); }

    /**
     * Return replacement value for named entity.
     * If entity is not found, empty string is returned.
     * @param name const char *, entity name
     */
   CString getReplacement(const char *name) const;

    /**
     * Add entity to map, If entity named 'name' exists already in map,
     * it's value is replaced with 'replacement'
     * @param name entity to add/change
     * @param replacement value that represents entity
     */
   void setEntity(const char *name, const char *replacement) { m_entities.setEntity(name, replacement); }

private:
   CXmlEntities m_entities;   ///< List of entities

   CString m_name;         ///< Document type name
   CString m_public_id;    ///< Public ID
   CString m_system_id;    ///< System ID

   bool m_html;            ///< HTML mode flag
};

/**
 * The CXmlDoc class represents the entire XML document.
 * It provides access to document root node, which includes all nodes in XML document tree.
 * 
 * Since elements, text nodes, comments, processing instructions, etc...
 * cannot exist outside the context of a document, the document class also contains 
 * the factory functions needed to create these objects. 
 *
 * The node objects created have an document() function which returns document which is used to create it.
 * Creation of elements, text nodes, etc... Is done using the various factory functions provided in this class.  
 */
class CXmlDoc: public CXmlNode
{
   friend class CXmlParser;
public:
    /**
     * Constructs an empty document, without doctype.
     */
   CXmlDoc();
    /**
     * Constructs an empty document, with doctype.
     * @param name const char *, name of the document.
     * @param public_id const char *, public id of the document, placed on DOCTYPE declaration
     * @param system_id const char *, system id of the document, placed on DOCTYPE declaration
     */
   CXmlDoc(const char *name, const char *public_id=0, const char *system_id=0);

    /**
     * Destroys object and free's all resources
     */
   virtual ~CXmlDoc() { clear(); }

    /**
     * Destroys all nodes in document
     */
   virtual void clear();

    /**
     * Creates new named node of type CXmlNode::DOM_ELEMENT.
     * It can be added to document DOM tree.
     * @param tagname const char *, name of the element
     * @see CXmlNode
     */
   CXmlNode *createElement(const char *tagname);

    /**
     * Creates new named node of type CXmlNode::DOM_TEXT.
     * It can be added to document DOM tree.
     * @param data const char *, content of the text node
     * @see CXmlNode
     */
   CXmlNode *createText(const char *data);

    /**
     * Creates new named node of type CXmlNode::DOM_COMMENT.
     * It can be added to document DOM tree.
     * @param data const char *, content of the comment node
     * @see CXmlNode
     */
   CXmlNode *createComment(const char *data);

    /**
     * Creates new named node of type CXmlNode::DOM_CDATA_SECTION.
     * It can be added to document DOM tree.
     * @param data const char *, content of CDATA section node
     * @see CXmlNode
     */
   CXmlNode *createCDataSection(const char *data);

    /**
     * Creates new named node of type CXmlNode::DOM_PI.
     * It can be added to document DOM tree.
     * @param target const char *, target of processing instruction
     * @param data const char *, data of processing instruction
     * @see CXmlNode
     */
   CXmlNode *createPI(const char *target, const char *data);

    /**
     * Returns doctype of document.
     * You can use it to add e.g. custom entities.
     <pre>
     mydoc->doctype().set_entity("myentity", "myreplacement");
     </pre>
     */
   CXmlDocType &docType() { return m_doctype; }

    /**
     * Returns doctype of document.
     * You can use it to add e.g. custom entities.
     <pre>
     mydoc->doctype().set_entity("myentity", "myreplacement");
     </pre>
     */
   const CXmlDocType &docType() const { return m_doctype; }

    /**
     * Returns pointer to root element of document.
     */
   CXmlNode *rootNode();

    /**
     * Returns indentation in save.
     */
   static int indentSpaces() { return m_indent_spaces; }

    /**
     * Set indentation in save, defaults to 2
     * @param i as new indent spaces
     */
   static void indentSpaces(int i) { m_indent_spaces = i; }

    /**
     * Save document to buffer.
     * @param buffer as CBuffer to save document
     * @param indent how many indent spaces at start
     */
   void save(CBuffer &buffer, int indent=0) const;

private:
   CXmlDocType m_doctype;        ///< Document type
   static int  m_indent_spaces;  ///< Indent spaces
};

/// @}

#endif
