/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CXmlNode.h  -  description
                             -------------------
    begin                : Sun May 22 2003
    based on the code    : Mikko Lahteenmaki <Laza@Flashmail.com>
    copyright            : (C) 2003 by Alexey S.Parshin
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
   This library is free software; you can redistribute it and/or modify it 
   under the terms of the GNU Library General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at 
   your option) any later version.

   This library is distributed in the hope that it will be useful, but 
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library
   General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; if not, write to the Free Software Foundation, 
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.

   Please report all bugs and problems to "alexeyp@gmail.com"
 ***************************************************************************/

#ifndef __CXMLNODE_H__
#define __CXMLNODE_H__

#include <sptk/CBuffer.h>
#include <sptk/CString.h>
#include <sptk/CStringMap.h>
#include <sptk/xml/CXmlNodeList.h>

class CXmlDoc;
class CXmlNode;

/// @addtogroup XML
/// @{

/**
 * The CXmlAttributes class is map for node attributes.
 */
class CXmlAttributes : public CStringMap<CString> {
   static CString emptyString;   ///< Empty string
public:
   typedef CStringMapIterator<CString> Iterator;   ///< Attributes iterator

   /// Constructs empty attribute map
   CXmlAttributes() { }

   /// Assign operator. Makes deep copy of map to another.
   /// @param src as copy source
   CXmlAttributes& operator = (const CXmlAttributes& src);

   /// Searches for named attribute.
   /// Returns true, if given attribute is found.
   /// @param attr CString, name of attribute to search
   bool hasAttribute(CString attr) const { return contains(attr.c_str()); }

   /// Returns attribute value for given attribute.
   /// If not found, empty string is returned.
   /// HTML tags can have empty attributes, for those you should use has_attribute() method.
   /// @param attr CString, name of attribute
   const CString &getAttribute(CString attr) const { 
      CString *st = find(attr.c_str());
      if (st) return *st;
      return CXmlAttributes::emptyString; 
   }

   /// Sets new value to attribute 'attr'.
   /// If attribute is not found, it's added to map.
   /// @param attr attribute name
   /// @param value attribute value
   void setAttribute(CString attr,const char *value) { insert(attr,value); }

   /// Sets new value to attribute 'attr'.
   /// If attribute is not found, it's added to map.
   /// @param attr attribute name
   /// @param value attribute value
   void setAttribute(CString attr, const CString &value) { insert(attr,value); }
};

/// XML node
class CXmlNode
{    
public:
   /// Node type enumeration
   enum NodeTypes {     
      DOM_DOCUMENT = 1,    /**< Document node */      
      DOM_ELEMENT,         /**< Normal element node, can contain subnodes */
      DOM_PI,              /**< Processing Instruction node */
      DOM_TEXT,            /**< Cdata where all default entities MUST be escaped. */
      DOM_CDATA_SECTION,      /**< Cdata section, which can contain preformatted char data. */
      DOM_COMMENT          /**< Comment node */
   };

   /// Destructor
   virtual ~CXmlNode();

   /// Deletes all children.
   virtual void clear();

   /**
    * Returns parent node of this node.
    * For CXmlDocument this returns NULL.
    */
   CXmlNode *parent() const { return m_parent; }

   /** 
    * Returns document context associated with this node.
    */
   CXmlDoc *document() const { return m_document; }

   /** 
    * Returns type of node.
    */
   int type() const { return m_nodetype; }

   /// Returns true node is document type 
   bool isDocument()      const { return (type()==DOM_DOCUMENT);  }

   /// Returns true node is element type
   bool isElement()    const { return (type()==DOM_ELEMENT); }

   /// Returns true node is element type, but has no children */
   bool isLeaf()       const { return (type()==DOM_ELEMENT) && m_child_nodes.count() == 0; }

   /// Returns true node is processing instruction (PI) type
   bool isPI()         const { return (type()==DOM_PI); }

   /// Returns true node is comment type
   bool isComment()    const { return (type()==DOM_COMMENT); }

   /// Returns true node is text type
   bool isText()       const { return (type()==DOM_TEXT); }

   /// Returns true node is cdata-section type
   bool isCDataSection() const { return (type()==DOM_CDATA_SECTION); }

   /** 
    * Returns the node name.
     * The meaning of the value depends on the node type: 
    <table border=0 bgcolor=#e8e8e8>
     <tr><td>DOM_DOCUMENT</td>      <td>#document</td>
     <tr><td>DOM_ELEMENT</td>    <td>The tag name</td>
     <tr><td>DOM_PI</td>         <td>The target of the processing instruction</td>
     <tr><td>DOM_COMMENT</td>    <td>#comment</td>
     <tr><td>DOM_TEXT</td>       <td>#text</td>
     <tr><td>DOM_CDATA_SECTION</td><td>#cdata-section</td>
    </table>
   */
   const CString &name() const { return m_name; }

   /// Sets the new name for the node
   /// @param name const CString &, new node name
   void name(const CString &name) { m_name = name; }

   /// Sets new name for node
   /// @param name const char *, new node name
   void name(const char *name) { m_name = name; }

   /**
    * Returns the value of the node.
    * The meaning of the value depends on the node type:
    <table border=0 bgcolor=#e8e8e8>
     <tr><td>DOM_DOCUMENT</td>      <td>none</td>
     <tr><td>DOM_ELEMENT</td>    <td>none</td>
     <tr><td>DOM_PI</td>         <td>The data of the processing intruction</td>
     <tr><td>DOM_COMMENT</td>    <td>The comment</td>
     <tr><td>DOM_TEXT</td>       <td>The text</td>
     <tr><td>DOM_CDATA_SECTION</td><td>The data of the cdata-section</td>
    </table>
    */
   CString &value() { return m_value; }

   /**
    * Returns the value of the node.
    * The meaning of the value depends on the node type:
    */
   const CString &value() const { return m_value; }

   /// Sets new value to node.
   /// @param new_value const CString &, new value
   /// @see value()
   void value(const CString &new_value) { m_value = new_value; }

   /// Sets new value to node.
   /// @param new_value const char *, value to set
   /// @see value()
   void value(const char *new_value) { m_value = new_value; }

    /**
    * Returns cdatas combined from children.
     * e.g. "Some <tag>text</tag> here" becomes: "Some text here"
    * @param ret CString &, string is stored here
    */
   void text(CString &ret);

    /**
    * Returns the first occurence of child with the given name.
     * Recursive is true, it will search throught children.
    * @param name const char *, name to be searched
    * @param recursive bool, recursive search mode
    */
   CXmlNode *child(const char *name, bool recursive=true) const;

    /**
    * Returns list of all subnodes with give name.
    * @param name to add list
    */
   CXmlNodeList nodes(const char *name);

   /**
    * Creates a deepcopy of the node. 
    * If deep is true, then the cloning is done recursively which means that all the 
    * node's children are deep copied too. If deep is false only the node itself is copied and the copy will have no child nodes. 
    */
   CXmlNode *cloneNode(bool deep=false);

   /// Removes the child node indicated by 'old_child' 
   /// from the list of children, and returns it. 
   /// @param child CXmlNode *, child node to remove
   CXmlNode *remove(CXmlNode *child);

   /// Replaces the child node 'old_child' with 'new_child'
   /// in the list of children, and returns the 'old_child' node. 
   /// If the 'new_child' is already in the tree, it is first removed. 
   /// @param newChild CXmlNode *, new child node
   /// @param oldChild CXmlNode *, old child node
   CXmlNode *replace(CXmlNode *newChild, CXmlNode *oldChild);

   /// Adds the node 'newChild' to the end of the list 
   /// of children of this node. If the 'new_child' is 
   /// already in the tree, it is first removed. 
   /// @param newChild CXmlNode *, new child node
   CXmlNode *add(CXmlNode *newChild);

   /**
    * Inserts the node 'new_child' before the existing child node 'ref_child'.
    * If 'ref_child' is null, insert 'new_child' at the end of the list of children.
    * If the 'new_child' is already in the tree, it is first removed. 
    * Returns the node being inserted. 
    */
   CXmlNode *insertBefore(CXmlNode *new_child, CXmlNode *ref_child);                                         

   /**
    * Returns list of child nodes.
    */
   const CXmlNodeList &childNodes() const { return m_child_nodes; }

   /**
    * Returns number of child nodes
    */
   unsigned children() const { return m_child_nodes.count(); }

   /**
    * Return child for given index, NULL if out of range.
    */
   CXmlNode *child(int index) const { return m_child_nodes.item(index); }

   /**
    * Returns reference to attribute map.
    * @see CXmlAttributes
    */
   const CXmlAttributes &attributes() const { return m_attributes; }

   /**
    * Returns true, if node has any attributes
    */
   bool hasAttributes() const { return (m_attributes.count()>0); }

   /**
    * Returns true, if given attribute is found
    * @param attr to search
    */
   bool hasAttribute(const char *attr) const { return m_attributes.hasAttribute(attr); }

   /**
    * Returns attribute value for given attribute.
    * If not found, empty string is returned. 
    * HTML tags can have empty attributes, for those you should use has_attribute() method.
    * @param attr name of attribute
    */
   const CString &getAttribute(CString attr) const { return m_attributes.getAttribute(attr); }

   /** 
    * Sets new value to attribute 'attr'.
    * If attribute is not found, it's added to map.
    * @param attr attribute name
    * @param value attribute value
    */
   void setAttribute(const char *attr, const char *value) { m_attributes.setAttribute(attr, value); }

   /** 
    * Sets new value to attribute 'attr'.
    * If attribute is not found, it's added to map.
    * @param attr attribute name
    * @param value attribute value
    */
   void setAttribute(const char *attr, const CString &value) { m_attributes.setAttribute(attr, value); }

    /** 
    * Saves node to buffer.
    * @param buffer to save
    * @param indent how many indent spaces at start
    */
   virtual void save(CBuffer &buffer, int indent=0) const;

protected:

   CString m_name;               ///< Node name
   CString m_value;              ///< Node value
   CXmlAttributes m_attributes;     ///< Node attributes

   unsigned char m_nodetype;     ///< Node type

   CXmlDoc *m_document;          ///< Document pointer

   CXmlNode *m_parent;           ///< Parent node pointer

   CXmlNodeList m_child_nodes;   ///< List of subnodes

private:
   friend class CXmlParser;
   friend class CXmlDoc;

   /// Sets parent node - for CXmlParser
   void parent(CXmlNode *p) { m_parent = p; }

   /// Sets node type
   void type(int type) { m_nodetype = (unsigned char)type; }

   /// Sets parent document
   CXmlNode(CXmlDoc *doc); 

   /// Constructor - for CXmlDoc
   CXmlNode(const CString &name, int type, CXmlDoc *doc) { m_name = name; m_nodetype = (unsigned char)type; m_document = doc; m_parent = 0; }

   //CXmlNode(const CXmlNode &node) { }
   //CXmlNode& operator = (const CXmlNode& s) { return *this; }
};

/// @}

#define XML_TYPE_DOC             CXmlNode::DOM_DOCUMENT
#define XML_TYPE_ELEMENT         CXmlNode::DOM_ELEMENT
#define XML_TYPE_PI              CXmlNode::DOM_PI
#define XML_TYPE_TEXT            CXmlNode::DOM_TEXT
#define XML_TYPE_CDATA_SECTION   CXmlNode::DOM_CDATA_SECTION
#define XML_TYPE_COMMENT         CXmlNode::DOM_COMMENT

#endif
