/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CBridgedDatabase.cpp  -  description
                             -------------------
    begin                : Jan 06 2004
    copyright            : (C) 1999-2005 by Alexey Parshin
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
   This library is free software; you can redistribute it and/or modify it 
   under the terms of the GNU Library General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at 
   your option) any later version.

   This library is distributed in the hope that it will be useful, but 
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library
   General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; if not, write to the Free Software Foundation, 
   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.

   Please report all bugs and problems to "alexeyp@gmail.com"
 ***************************************************************************/

#include <stdio.h>
#include <sptk/CBridgedDatabase.h>
#include <sptk/CQuery.h>
#include <sptk/CException.h>
#include "CODBCChannel.h"

CBridgedDatabase::CBridgedDatabase(CString hostName,CString connectionString)
: CDatabase(connectionString)
{
   m_active = false;
   m_inTransaction = false;
   m_socket.host("localhost");
   m_socket.port(3000);
   m_odbcChannel = new CODBCChannel(&m_socket);
}

CBridgedDatabase::~CBridgedDatabase()
{
   puts("CBridgedDatabase::~CBridgedDatabase()");
   try {
      if (m_socket.active()) 
         close();
      delete m_odbcChannel;
   }
   catch (...) {
   }
}

static inline bool successful(int ret) {
   // 0 and 1 only
   return (ret & 1) == ret;
}

CString CBridgedDatabase::readError() {
   char errorText[2048];
   m_odbcChannel->readData(errorText,2048);
   return errorText;
}

void CBridgedDatabase::open(CString connectionString)
{
   if (connectionString.length())
      m_connString = connectionString;

   m_active = false;

   m_socket.open();
   m_odbcChannel->beginSend(OCT_DB_OPEN);
   m_odbcChannel->sendData(m_connString);
   m_odbcChannel->commitSend();
   int rc = m_odbcChannel->readResponse();
   if (!successful(rc))
      throw CException(readError());
   m_active = true;
}

void CBridgedDatabase::close() {
   if (m_active) {
      puts("CBridgedDatabase::close()");

      if (m_inTransaction)
         rollbackTransaction();

      for (unsigned i = 0; i < m_queryList.count(); i++) {
         CQuery *query = (CQuery *)m_queryList[i];
         queryFreeStmt(query);
      }

      m_inTransaction = false;

      int rc = 0;

      if (m_active) {
         m_odbcChannel->beginSend(OCT_DB_CLOSE);
         m_odbcChannel->commitSend();
         rc = m_odbcChannel->readResponse();
      }

      puts("Database closed, closing socket");

      m_active = false;

      if (!successful(rc))
         throw CException(readError());

      m_socket.close();

      puts("Socket closed");

   }
}

CString CBridgedDatabase::queryError(const CQuery* query) const
{
   m_odbcChannel->beginSend(OCT_DB_ERRORINFO);
   m_odbcChannel->commitSend();
   CString errorInfo;
   m_odbcChannel->readData(errorInfo);
   return errorInfo;
}

bool CBridgedDatabase::active() const
{
   return m_active;
}

void CBridgedDatabase::queryFreeStmt(CQuery* query)
{
   puts("CBridgedDatabase::queryFreeStmt()");
   m_odbcChannel->beginSend(OCT_QUERY_FREESTMT);
   m_odbcChannel->sendData((int)queryHandle(query));
   m_odbcChannel->commitSend();
   int rc = m_odbcChannel->readResponse();
   if (!successful(rc))
      throw CException(readError());
   querySetStmt(query,0L);
   puts("CBridgedDatabase::queryFreeStmt completed");
}

void CBridgedDatabase::beginTransaction()
{
   m_odbcChannel->beginSend(OCT_DB_BEGINTRAN);
   m_odbcChannel->commitSend();
   int rc = m_odbcChannel->readResponse();
   if (!successful(rc))
      throw CException(readError());
   m_inTransaction = true;
}

void CBridgedDatabase::endTransaction(bool commit)
{
   m_odbcChannel->beginSend(OCT_DB_ENDTRAN);
   m_odbcChannel->sendData(int(commit));
   m_odbcChannel->commitSend();
   int rc = m_odbcChannel->readResponse();
   if (!successful(rc))
      throw CException(readError());
   m_inTransaction = false;
}

void CBridgedDatabase::querySetParameters(CQuery *query) {
   unsigned paramCount = query->paramCount();
   if (!paramCount) return;
   m_odbcChannel->beginSend(OCT_QUERY_SETPARAMS);
   m_odbcChannel->sendData((int)query);
   m_odbcChannel->sendData((int)paramCount);
   for (unsigned i = 0; i < paramCount; i++) {
      CParam& param = query->param(i);
      switch (param.dataType()) {
         case VAR_INT:
            m_odbcChannel->sendData('I');
            m_odbcChannel->sendData(param.getInteger());
            break;
         case VAR_DATE:
            m_odbcChannel->sendData('D');
            m_odbcChannel->sendData(param.getInteger());
            break;
         case VAR_FLOAT:
            m_odbcChannel->sendData('F');
            m_odbcChannel->sendData(param.getFloat());
            break;
         case VAR_DATE_TIME:
            m_odbcChannel->sendData('T');
            m_odbcChannel->sendData(param.getFloat());
            break;
         case VAR_STRING:
            m_odbcChannel->sendData('S');
            m_odbcChannel->sendData(param.getString());
            break;
         case VAR_TEXT:
         case VAR_BUFFER:
            m_odbcChannel->sendData('B');
            m_odbcChannel->sendData(param.getString(),param.dataSize());
            break;
         default:
            m_odbcChannel->sendData('U');
            puts("Unsupported parameter type");
            break;
      }
      if (param.isNull())
         m_odbcChannel->sendData('Y');
      else m_odbcChannel->sendData('N');
   }
   m_odbcChannel->commitSend();
   int rc = m_odbcChannel->readResponse();
   if (!successful(rc))
      throw CException(readError());
}

void CBridgedDatabase::queryBindParameters(CQuery* query)
{
   CDatabase::queryBindParameters(query);
}

void CBridgedDatabase::queryOpen(CQuery* query)
{
   m_odbcChannel->beginSend(OCT_QUERY_OPEN);
   m_odbcChannel->sendData((int)queryHandle(query));
   m_odbcChannel->sendData(query->sql());
   m_odbcChannel->commitSend();
   int rc = m_odbcChannel->readResponse();
   if (!successful(rc))
      throw CException(readError());
   int handle;
   m_odbcChannel->readData(handle);
   queryHandle(query,(void *)handle);
}

void CBridgedDatabase::queryCloseStmt(CQuery* query)
{
   m_odbcChannel->beginSend(OCT_QUERY_CLOSE);
   m_odbcChannel->sendData((int)queryHandle(query));
   m_odbcChannel->commitSend();
   int rc = m_odbcChannel->readResponse();
   if (!successful(rc))
      throw CException(readError());
}

void CBridgedDatabase::queryColAttributes(CQuery* query, short column, short descType, char* buff, int len)
{
   CDatabase::queryColAttributes(query, column, descType, buff, len);
}

void CBridgedDatabase::queryColAttributes(CQuery* query, short column, short descType, long& value)
{
   CDatabase::queryColAttributes(query, column, descType, value);
}

void CBridgedDatabase::queryExecute(CQuery* query)
{
   CDatabase::queryExecute(query);
}

void CBridgedDatabase::queryFetch(CQuery* query)
{
   CDatabase::queryFetch(query);
}

void CBridgedDatabase::queryPrepare(CQuery* query)
{
   CDatabase::queryPrepare(query);
}

void CBridgedDatabase::queryUnprepare(CQuery* query)
{
   CDatabase::queryUnprepare(query);
}
